; Protected-Mode code & data for the TUNES BIOS-based API
; see ~/info/BIOS-API.nfo

pmpart

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Constants ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
Orig_INT_NUM	=	0x10	; first interrupt called



__dataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Global Variables ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

rm_idt_address:
		.word	256*4-1	; Limit of real-mode IDT
		.long	0	; Base Address of real-mode IDT

pm_idt_address:
		.word	256*8-1	; Limit of protected-mode IDT
		.long	0 ; ?	; Base Address of protected-mode IDT

XtdMem:		.long	0
ConvMem:	.long	0
AvConvMem:	.long	0
CPUType:	.word	0

RM_Part_ofs:	.word	rmpart_beg	; RM offset to RM part
RM_Part_seg:	.word	0 ; ?		; RM segment for RM part

})


__udataseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;; Uninitialized Data ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

RM_Part_PA	= allot(u,4)	; physical address of real-mode part
PM_Part_PA	= allot(u,4)	; physical address of protected-mode part
KERNEL_PA	= allot(u,4)	; physical address of Kernel.
RM_Part_Addr	= allot(u,0)
PM_SYSTACK	= allot(u,3*2)	; Saved address for P-mode stack
PM_LDT_sel	= allot(u,2)	; Address of protected-mode LDT
CPU_Type	= allot(u,2)	; What processor/coprocessor this is.

__GDT__		= alallot(u,16,8*GDT_SIZE)
__IDT__		= alallot(u,16,8*IDT_SIZE)
__Stack__	= alallot(u,16,4*PM_STACK_SIZE)
__Stack_Top__	= allot(u,0)
})





__codeseg__({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;; Hardware IO modes ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

Set_MOOSE_IO_mode:
;;; program I/O devices for MOOSE.
; PREconditions:
; * Interrupt,Global,Local,Page Tables have been initialized
;   IRQ's and NMI's are disabled (were disabled before RM->PM switching)
; POSTconditions:
; * IDT,GDT,LDT,TSS,CR3 registers are loaded.
;
  ;;; disable IRQ's
	cli
  ;;; disable NMI's
	call Disable_NMI
  ;;; load protected-mode idt
	lidt [pmaddr(pm_idt_address)]


Set_BIOS_IO_mode:
;;; reprogram I/O devices so the BIOS can use them
; Must be the exact converse of Set_MOOSE_IO_mode
;
; currently, reenable all interrupts
   ;;; reenable NMI's
	call Enable_NMI
   ;;; reenable IRQ's
	sti
   ;;; return to the caller
	ret

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; I/O subroutines ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

define({__IO_Delay__},
	{locals({D1},{D2},{jmp D1}__nl__{D1:	jmp D2}__nl__{D2:})})

LibProc(Disable_NMI)
LibProc(Enable_NMI)


;;;;;;;;;;;;;;;;;;;;;;;;;; Initialization routines ;;;;;;;;;;;;;;;;;;;;;;;;;;;

Setup_PM_8259:
;; Taken from Linux' boot/setup.S
 	mov	al,#0x11	; initialization sequence
	out	0x20,al		; send it to 8259A-1
	__IO_Delay__
	out	0xA0,al		; and to 8259A-2
	__IO_Delay__
	mov	al,#IRQ_INT_BASE	; start of hardware int's (0x90)
	out	0x21,al
	__IO_Delay__
	mov	al,#IRQ_INT_BASE+8	; start of hardware int's 2 (0x98)
	out	0xA1,al
	__IO_Delay__
	mov	al,0x4		; 8259-1 is master
	out	0x21,al
	__IO_Delay__
	mov	al,0x2		; 8259-2 is slave
	out	0xA1,al
	__IO_Delay__
	mov	al,0x1		; 8086 mode for both
	out	0x21,al
	__IO_Delay__
	out	0xA1,al
	__IO_Delay__
;; not needed: already all interrupts masked off.
;	mov	al,0xFF		; mask off all interrupts for now
;	out	0xA1,al
;	__IO_Delay__
;	mov	al,0xFB		; mask all irq's but irq2 which
;	out	0x21,al			; is cascaded
	ret

Setup_RM_8259:
	mov	al,0x11		; initialization sequence
	out	0x20,al		; send it to 8259A-1
	__IO_Delay__
	out	0xA0,al		; and to 8259A-2
	__IO_Delay__
	mov	al,0x8		; start of hardware int's (0x08 argl !)
	out	0x21,al
	__IO_Delay__
	mov	al,0x70		; start of hardware int's 2 (0x70)
	out	0xA1,al
	__IO_Delay__
	mov	al,0x4		; 8259-1 is master
	out	0x21,al
	__IO_Delay__
	mov	al,0x2		; 8259-2 is slave
	out	0xA1,al
	__IO_Delay__
	mov	al,0x1		; 8086 mode for both
	out	0x21,al
	__IO_Delay__
	out	0xA1,al
	__IO_Delay__
;; not needed: already all interrupts masked off.
;	mov	al,0xFF		; mask off all interrupts for now
;	out	0xA1,al
;	__IO_Delay__
;	mov	al,0xFB		; mask all irq's but irq2 which
;	out	0x21,al			; is cascaded
	ret


;; enable A20, reset copro, reprogram interrupts...
;; see Linux: boot/setup.S


;;;;;;;;;;;;;;;;;;;;;;; Protected to Real Mode switcher ;;;;;;;;;;;;;;;;;;;;;;;

RM_INT:	; FORTH ( TSS NUM -- )
; Parameters:
;  TSS is a TSS-like record of registers
;  NUM is the number of the real mode interrupt to call
; Action:
;  Interrupt number NUM is called in real mode with
;  register values from TSS; values in TSS are updated
;  according to result from interrupt.
;  CS:EIP field of the TSS is broken.
; Calling convention:
;  DS points on pmpart
;  SS:ESP must be valid (no RET-based threading)

   ;;; modify code to set interrupt number to call
	push #RMPARTDS
	pop fs
	;; (!) beware if offset on RMPART segment
	seg fs
	mov [rmaddr(INT_NUM)],bl
   ;;; set jump address in TSS to that of DoCallINT
	mov ebx,[edi]		; pop NUM
	add edi,#4
	;; (!) beware if offset on RMPART segment
	mov [ebx+TSS32_EIP],#rmaddr(DoCallINT)
	mov ax,[pmaddr(RM_Part_seg)]
	mov [ebx+TSS32_CS],ax
   ;;; Jump into real-mode
	;jmp Go_Real_TSS

Go_Real_TSS:
   ;;; save Protected mode registers
	pushm(ad,ds,es,fs,gs)
	mov [pmaddr(PM_SYSTACK)],esp
	mov [pmaddr(PM_SYSTACK+4)],ss

   ;;; set parameters for Real-Mode call
	push #RMPARTDS
	pop fs
	seg fs
	lss esp,[rmaddr(RM_SYSTACK)]
	seg fs
	mov dword [rmaddr(PM_Target_ofs)],#pmaddr(Ret_From_RM_INT)
	seg fs
	mov word [rmaddr(PM_Target_seg)],#PMPARTCS
	;jmp Go_BIOS_Real

Go_BIOS_Real:
; PREconditions:
; * 
; POSTconditions:
; * Restore Interrupts, I/O devices, and BIOS Data area,
;  so that the BIOS will run properly
; * Do the switching to Real mode
; * jump to address pointed by [RM_Target]

	lidt [pmaddr(rm_idt_address)]	; beware segment restrictions !
_GoReal:
	jmpi GoReal, RMPARTCS


Ret_From_RM:
	call Enable_NMI

Ret_From_RM_INT:
	seg cs
	lss esp,[pmaddr(PM_SYSTACK)]
	popm(ad,ds,es,fs,gs)
	ret
})














forget({
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; IO Functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
	;;;; Buffered BIOS-based Console output ;;;;
__udataseg__({

; Buffer for Buffered BIOS-based Console output.
define({BBO_BUFSIZE},	256)
BBO_Buffer	= allot(u,BBO_BUFSIZE)
})

__dataseg__({
BBO_BP		.long	0
})

__codeseg__({
BBPutc:
	pushad
	mov edi,[pmaddr(BBO_BP)]
	cld
	stosb
	__IF__(ae)
	  cmp edi,#BBO_BUFSIZE
	__THEN__
	  call BBOFlush
	__ENDIF__
	popad
	ret

BBOFlush:
	;mov eax,[rmaddr(__PD__)]
	;push #rmaddr(RM_BBOFlush)
	jmp Go_BIOS_Real
})
})
